@@ -35,6 +35,7 @@
 static int ifxmodem_init(void)
 {
 	ifx_voicecall_init();
+	ifx_audio_settings_init();
 	ifx_radio_settings_init();
 	ifx_stk_init();
 
@@ -45,6 +46,7 @@ static void ifxmodem_exit(void)
 {
 	ifx_stk_exit();
 	ifx_radio_settings_exit();
+	ifx_audio_settings_exit();
 	ifx_voicecall_exit();
 }
 
